#
# Copyright 2020 Joachim Lusiardi
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import unittest
import enum

import tlv8


class TestTLV8Enum(unittest.TestCase):
    def test_decode_enum(self):

        class TestKeys(enum.IntEnum):
            KEY_1 = 1
            KEY_2 = 2

        class TestValues(enum.IntEnum):
            VALUE_1 = 1
            VALUE_2 = 2

        data = b'\x01\x01\x02'
        result = tlv8.decode(data, {
            TestKeys.KEY_1: TestValues
        })
        expected = tlv8.EntryList([
            tlv8.Entry(TestKeys.KEY_1, TestValues.VALUE_2)
        ])
        self.assertIsInstance(result[0].type_id, TestKeys)
        self.assertIsInstance(result[0].data, TestValues)
        self.assertEqual(expected, result)

    def test_decode_key(self):
        class TestKeys(enum.IntEnum):
            KEY_1 = 1
            KEY_2 = 2
        data = b'\x01\x03foo'
        result = tlv8.decode(data, {
            TestKeys.KEY_1: tlv8.DataType.STRING
        })
        expected = tlv8.EntryList([
            tlv8.Entry(TestKeys.KEY_1, 'foo')
        ])
        self.assertIsInstance(result[0].type_id, TestKeys)
        self.assertEqual(expected, result)

    def test_decode_value(self):
        class TestValues(enum.IntEnum):
            VALUE_1 = 1
            VALUE_2 = 2
        data = b'\x01\x01\x02'
        result = tlv8.decode(data, {
            1: TestValues
        })
        expected = tlv8.EntryList([
            tlv8.Entry(1, 2)
        ])
        self.assertIsInstance(result[0].data, TestValues)
        self.assertEqual(expected, result)

    def test_encode_key(self):
        class TestKeys(enum.IntEnum):
            KEY_1 = 1
            KEY_2 = 2
        data = [
            tlv8.Entry(TestKeys.KEY_1, 'foo')
        ]
        result = tlv8.encode(data)
        self.assertEqual(b'\x01\x03foo', result)

    def test_encode_int_value(self):
        class TestValues(enum.IntEnum):
            VAL_1 = 1
            VAL_2 = 2

        data = [
            tlv8.Entry(1, TestValues.VAL_2)
        ]
        result = tlv8.encode(data)
        self.assertEqual(b'\x01\x01\x02', result)



#
# Copyright 2020 Joachim Lusiardi
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import unittest
from struct import pack

import tlv8


class TestTLV8DecodeInteger(unittest.TestCase):
    def test_decode_int1_neg(self):
        input_data = b'\x01\x01' + pack('<b', -123)
        structure = {
            1: tlv8.DataType.INTEGER,
        }
        result = tlv8.decode(input_data, structure)
        expected = tlv8.EntryList([
            tlv8.Entry(1, -123),
        ])
        self.assertEqual(expected, result)

    def test_decode_int1_pos(self):
        input_data = b'\x01\x01' + pack('<b', 123)
        structure = {
            1: tlv8.DataType.INTEGER,
        }
        result = tlv8.decode(input_data, structure)
        expected = tlv8.EntryList([
            tlv8.Entry(1, 123),
        ])
        self.assertEqual(expected, result)

    def test_decode_int1_un(self):
        input_data = b'\x01\x01' + pack('<B', 123)
        structure = {
            1: tlv8.DataType.UNSIGNED_INTEGER,
        }
        result = tlv8.decode(input_data, structure)
        expected = tlv8.EntryList([
            tlv8.Entry(1, 123),
        ])
        self.assertEqual(expected, result)

    def test_decode_int2_neg(self):
        input_data = b'\x01\x02' + pack('<h', -12345)
        structure = {
            1: tlv8.DataType.INTEGER,
        }
        result = tlv8.decode(input_data, structure)
        expected = tlv8.EntryList([
            tlv8.Entry(1, -12345),
        ])
        self.assertEqual(expected, result)

    def test_decode_int2_pos(self):
        input_data = b'\x01\x02' + pack('<h', 12345)
        structure = {
            1: tlv8.DataType.INTEGER,
        }
        result = tlv8.decode(input_data, structure)
        expected = tlv8.EntryList([
            tlv8.Entry(1, 12345),
        ])
        self.assertEqual(expected, result)

    def test_decode_int2_un(self):
        input_data = b'\x01\x02' + pack('<H', 12345)
        structure = {
            1: tlv8.DataType.UNSIGNED_INTEGER,
        }
        result = tlv8.decode(input_data, structure)
        expected = tlv8.EntryList([
            tlv8.Entry(1, 12345),
        ])
        self.assertEqual(expected, result)

    def test_decode_int3_signed(self):
        input_data = b'\x01\x03\x01\x02\x03\x02\x02\x02\x03'
        structure = {
            1: tlv8.DataType.INTEGER,
        }
        self.assertRaises(ValueError, tlv8.decode, input_data, structure)

    def test_decode_int3_unsigned(self):
        input_data = b'\x01\x03\x01\x02\x03\x02\x02\x02\x03'
        structure = {
            1: tlv8.DataType.UNSIGNED_INTEGER,
        }
        self.assertRaises(ValueError, tlv8.decode, input_data, structure)
        # tlv8.decode(input_data, structure)

    def test_decode_int4_neg(self):
        input_data = b'\x01\x04' + pack('<i', -12345)
        structure = {
            1: tlv8.DataType.INTEGER,
        }
        result = tlv8.decode(input_data, structure)

        expected = tlv8.EntryList([
            tlv8.Entry(1, -12345),
        ])
        self.assertEqual(expected, result)

    def test_decode_int4_pos(self):
        input_data = b'\x01\x04' + pack('<i', 12345)
        structure = {
            1: tlv8.DataType.INTEGER,
        }
        result = tlv8.decode(input_data, structure)

        expected = tlv8.EntryList([
            tlv8.Entry(1, 12345),
        ])
        self.assertEqual(expected, result)

    def test_decode_int4_un(self):
        input_data = b'\x01\x04' + pack('<I', 12345)
        structure = {
            1: tlv8.DataType.UNSIGNED_INTEGER,
        }
        result = tlv8.decode(input_data, structure)

        expected = tlv8.EntryList([
            tlv8.Entry(1, 12345),
        ])
        self.assertEqual(expected, result)

    def test_decode_int8_neg(self):
        input_data = b'\x01\x08' + pack('<q', -4611686018427387904)
        structure = {
            1: tlv8.DataType.INTEGER,
        }
        result = tlv8.decode(input_data, structure)

        expected = tlv8.EntryList([
            tlv8.Entry(1, -4611686018427387904),
        ])
        self.assertEqual(expected, result)

    def test_decode_int8_pos(self):
        input_data = b'\x01\x08' + pack('<q', 4611686018427387904)
        structure = {
            1: tlv8.DataType.INTEGER,
        }
        result = tlv8.decode(input_data, structure)

        expected = tlv8.EntryList([
            tlv8.Entry(1, 4611686018427387904),
        ])
        self.assertEqual(expected, result)

    def test_decode_int8_un(self):
        input_data = b'\x01\x08' + pack('<q', 4611686018427387904)
        structure = {
            1: tlv8.DataType.UNSIGNED_INTEGER,
        }
        result = tlv8.decode(input_data, structure)

        expected = tlv8.EntryList([
            tlv8.Entry(1, 4611686018427387904),
        ])
        self.assertEqual(expected, result)



#
# Copyright 2020 Joachim Lusiardi
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

__all__ = [
    'TestTLV8', 'TestTLV8Decode', 'TestTLV8Entry', 'TestTLV8Enum', 'TestTLV8EntryList', 'TestTLV8DeepDecode',
    'TestTLV8DecodeInteger', 'TestTLV8RealWorld', 'TestTLV8ToJson'
]

from tests.tlv8_encode_tests import TestTLV8
from tests.tlv8_decode_tests import TestTLV8Decode
from tests.tlv8_deep_decode_tests import TestTLV8DeepDecode
from tests.tlv8_entry_tests import TestTLV8Entry
from tests.tlv8_enum_test import TestTLV8Enum
from tests.tlv8_entrylist_tests import TestTLV8EntryList
from tests.tlv8_decode_integer_tests import TestTLV8DecodeInteger
from tests.tlv8_real_world_test import TestTLV8RealWorld
from tests.tlv8_to_json_test import TestTLV8ToJson



#
# Copyright 2020 Joachim Lusiardi
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import unittest
from struct import pack

import tlv8


class TestTLV8Decode(unittest.TestCase):
    def test_zero_input(self):
        input_data = b''
        result = tlv8.decode(input_data)
        self.assertEqual(tlv8.EntryList(), result)

    def test_input_to_short_1(self):
        input_data = b'\x01'
        self.assertRaises(ValueError, tlv8.decode, input_data)

    def test_input_to_short_2(self):
        input_data = b'\x01\x01'
        self.assertRaises(ValueError, tlv8.decode, input_data)

    def test_decode_single_entry(self):
        input_data = b'\x02\x01\x23'
        result = tlv8.decode(input_data)
        self.assertEqual(tlv8.EntryList([tlv8.Entry(2, b'\x23')]), result)

    def test_decode_single_entry_bytearry(self):
        input_data = bytearray(b'\x02\x01\x23')
        result = tlv8.decode(input_data)
        self.assertEqual(tlv8.EntryList([tlv8.Entry(2, b'\x23')]), result)

    def test_decode_2_entries(self):
        input_data = b'\x02\x01\x23\x03\x01\x42'
        result = tlv8.decode(input_data)
        self.assertEqual(tlv8.EntryList([tlv8.Entry(2, b'\x23'), tlv8.Entry(3, b'\x42')]), result)

    def test_decode_2_entries_1_expected(self):
        input_data = b'\x02\x01\x23\x03\x01\x42'
        structure = {
            2: tlv8.DataType.INTEGER
        }
        result = tlv8.decode(input_data, structure)
        self.assertEqual(tlv8.EntryList([tlv8.Entry(2, 0x23)]), result)

    def test_decode_4(self):
        input_data = b'\x01\x01\x23\x02\x03\x04\x01\x42\x01\x01\x23'
        structure = {
            1: tlv8.DataType.INTEGER,
            2: {
                4: tlv8.DataType.INTEGER,
            }
        }
        result = tlv8.decode(input_data, structure)
        expected = tlv8.EntryList([
            tlv8.Entry(1, 0x23),
            tlv8.Entry(2, tlv8.EntryList([
                tlv8.Entry(4, 0x42),
            ])),
            tlv8.Entry(1, 0x23),
        ])
        self.assertEqual(expected, result)

    def test_decode_float(self):
        input_data = b'\x01\x04' + pack('<f', 3.141)
        structure = {
            1: tlv8.DataType.FLOAT,
        }
        result = tlv8.decode(input_data, structure)
        expected = tlv8.EntryList([
            tlv8.Entry(1, 3.141),
        ])
        self.assertEqual(expected, result)

    def test_decode_example_2(self):
        input_data = b'\x01\x01\x00'
        structure = {
            1: tlv8.DataType.INTEGER,
        }
        result = tlv8.decode(input_data, structure)
        expected = tlv8.EntryList([
            tlv8.Entry(1, 0),
        ])
        self.assertEqual(expected, result)

    def test_decode_example_3(self):
        input_data = b'\x01\x15\x01\x10e\xad\x8b\xe8\xb3fD\xcb\xbde#\xccc\n\xb8\xef\x02\x01\x01'
        structure = {
            1: {
                1: tlv8.DataType.BYTES,
                2: tlv8.DataType.INTEGER,
            },
            2: {
                1: tlv8.DataType.INTEGER,
                2: tlv8.DataType.BYTES,
                3: tlv8.DataType.BYTES,
                4: tlv8.DataType.BYTES
            },
            3: {
                1: tlv8.DataType.INTEGER,
                2: tlv8.DataType.INTEGER,
                3: tlv8.DataType.INTEGER,
                4: tlv8.DataType.INTEGER,
                5: tlv8.DataType.INTEGER
            }
        }
        result = tlv8.decode(input_data, structure)
        expected = tlv8.EntryList([
            tlv8.Entry(1, tlv8.EntryList([
                tlv8.Entry(1, b'e\xad\x8b\xe8\xb3fD\xcb\xbde#\xccc\n\xb8\xef'),
                tlv8.Entry(2, 1),
            ])),
        ])
        self.assertEqual(expected, result)

    def test_decode_257bytes(self):
        data = b''
        for i in range(0, 257):
            data += pack('<B', i % 256)

        input_data = b'\x17\xff' + data[0:255] + b'\x17\x02' + data[255:]

        result = tlv8.decode(input_data)

        expected = tlv8.EntryList([
            tlv8.Entry(23, data)
        ])
        self.assertEqual(result, expected)

    def test_decode_supported_rtp_configs(self):
        data = \
            b'\x02\x01\x00' + \
            b'\x00\x00' + \
            b'\x02\x01\x01'
        result = tlv8.decode(data, expected={2: tlv8.DataType.INTEGER})
        expected_data = tlv8.EntryList([
            tlv8.Entry(2, 0),
            tlv8.Entry(2, 1)
        ])
        self.assertEqual(result, expected_data)

    def test_encode_supported_video_stream_configuration(self):
        data = \
            b'\x01\x30' \
                b'\x01\x01\x00' \
                b'\x02\x0f' \
                    b'\x01\x01\x00' \
                    b'\x02\x01\x00' \
                    b'\x03\x01\x00' \
                    b'\x04\x01\x01' \
                    b'\x05\x01\n' \
                b'\x03\x0b' \
                    b'\x01\x02\x00\x05' \
                    b'\x02\x02 \x03' \
                    b'\x03\x01\x1e' \
                b'\xff\x00' \
                b'\x03\x0b' \
                    b'\x01\x02\x80\x02' \
                    b'\x02\x02\xe0\x01' \
                    b'\x03\x01\x1e' \
            b'\xff\x00' \
            b'\x01\x2d' \
                b'\x01\x01\x00' \
                b'\x02\x0c' \
                    b'\x01\x01\x00' \
                    b'\x02\x01\x00' \
                    b'\x03\x01\x00' \
                    b'\x04\x01\x00' \
                b'\x03\x0b' \
                    b'\x01\x02\x00\x05' \
                    b'\x02\x02 \x03' \
                    b'\x03\x01\x1e' \
                b'\xff\x00' \
                b'\x03\x0b' \
                    b'\x01\x02\x80\x02' \
                    b'\x02\x02\xe0\x01' \
                    b'\x03\x01\x1e'
        expected_data_video_codec_params = {
            1: tlv8.DataType.INTEGER,
            2: tlv8.DataType.INTEGER,
            3: tlv8.DataType.INTEGER,
            4: tlv8.DataType.INTEGER,
            5: tlv8.DataType.INTEGER
        }
        expected_data_video_attribs = {
            1: tlv8.DataType.INTEGER,
            2: tlv8.DataType.INTEGER,
            3: tlv8.DataType.INTEGER
        }
        expected_data_video_codec_config = {
            1: tlv8.DataType.INTEGER,
            2: expected_data_video_codec_params,
            3: expected_data_video_attribs
        }
        expected_data_structure = {
            1: expected_data_video_codec_config
        }
        result = tlv8.decode(data, expected_data_structure)
        expected_data = tlv8.EntryList([
            tlv8.Entry(1, tlv8.EntryList([  # video codec config
                tlv8.Entry(1, 0),  # h.264
                tlv8.Entry(2, tlv8.EntryList([  # video codec param
                    tlv8.Entry(1, 0),  # Constrained Baseline Profile
                    tlv8.Entry(2, 0),  # Level 3.1
                    tlv8.Entry(3, 0),  # Packetization mode: Non-interleaved mode
                    tlv8.Entry(4, 1),  # CVO not enabled
                    tlv8.Entry(5, 10),  # CVO ID
                ])),
                tlv8.Entry(3, tlv8.EntryList([  # video attributes
                    tlv8.Entry(1, 1280),  # width
                    tlv8.Entry(2, 800),  # height
                    tlv8.Entry(3, 30)  # fps
                ])),
                tlv8.Entry(3, tlv8.EntryList([  # video attributes
                    tlv8.Entry(1, 640),  # width
                    tlv8.Entry(2, 480),  # height
                    tlv8.Entry(3, 30)  # fps
                ]))
            ])),
            tlv8.Entry(1, tlv8.EntryList([
                tlv8.Entry(1, 0),  # h.264
                tlv8.Entry(2, tlv8.EntryList([  # video codec param
                    tlv8.Entry(1, 0),  # Constrained Baseline Profile
                    tlv8.Entry(2, 0),  # Level 3.1
                    tlv8.Entry(3, 0),  # Packetization mode: Non-interleaved mode
                    tlv8.Entry(4, 0),  # CVO not enabled
                ])),
                tlv8.Entry(3, tlv8.EntryList([  # video attributes
                    tlv8.Entry(1, 1280),  # width
                    tlv8.Entry(2, 800),  # height
                    tlv8.Entry(3, 30)  # fps
                ])),
                tlv8.Entry(3, tlv8.EntryList([  # video attributes
                   tlv8.Entry(1, 640),  # width
                   tlv8.Entry(2, 480),  # height
                   tlv8.Entry(3, 30)  # fps
                ]))
            ]))
        ])

        self.assertEqual(result, expected_data)

    def test_decode_error_1(self):
        self.assertRaises(ValueError, tlv8.decode, {})

    def test_decode_missing_separator_strict(self):
        data = b'\x01\x01\x02\x01\x01\x02'
        self.assertRaises(ValueError, tlv8.decode, data, strict_mode=True)

    def test_decode_missing_separator_nonstrict(self):
        data = b'\x01\x01\x02\x01\x01\x02'
        result = tlv8.decode(data, strict_mode=False)
        self.assertEqual(result, tlv8.EntryList([tlv8.Entry(1, b'\x02'), tlv8.Entry(1, b'\x02')]))

    def test_decode_error_4(self):
        data = b'\x01\x02Hi'
        result = tlv8.decode(data, {1: tlv8.DataType.STRING})
        expected_data = tlv8.EntryList([
            tlv8.Entry(1, 'Hi')
        ])
        self.assertEqual(result, expected_data)

    def test_decode_error_5(self):
        data = b'\x01\x05\x01\x01\x01\x01\x01'
        structure = {1: tlv8.DataType.INTEGER}
        self.assertRaises(ValueError, tlv8.decode, data, structure)

    def test_decode_error_unknown_type(self):
        data = b'\x01\x05\x01\x01\x01\x01\x01'
        structure = {1: "string"}
        with self.assertRaises(ValueError) as error_context:
            tlv8.decode(data, structure)
        self.assertEqual(str(error_context.exception), 'Decoding failed, unknown data type: string')



#
# Copyright 2020 Joachim Lusiardi
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import unittest

import tlv8


class TestTLV8Entry(unittest.TestCase):
    def test_str_str(self):
        entry1 = tlv8.Entry(1, 'Hallo')
        self.assertEqual(entry1.__str__(), '<Entry 1, Hallo>')

    def test_str_bytes(self):
        entry1 = tlv8.Entry(1, b'23')
        self.assertEqual(entry1.__str__(), '<Entry 1, b\'23\'>')

    def test_str_int(self):
        entry1 = tlv8.Entry(1, 23)
        self.assertEqual(entry1.__str__(), '<Entry 1, 23>')

    def test_str_float(self):
        entry1 = tlv8.Entry(1, 23.42)
        self.assertEqual(entry1.__str__(), '<Entry 1, 23.42>')

    def test_equal_different_types(self):
        entry1 = tlv8.Entry(1, 42)
        self.assertNotEqual(entry1, 'ping')

    def test_equal_different_ids(self):
        entry1 = tlv8.Entry(1, 42)
        entry2 = tlv8.Entry(2, 42)
        self.assertNotEqual(entry1, entry2)

    def test_equal_different_floats(self):
        entry1 = tlv8.Entry(1, 42.1)
        entry2 = tlv8.Entry(1, 42.2)
        self.assertNotEqual(entry1, entry2)

    def test_equal_same_floats(self):
        entry1 = tlv8.Entry(1, 42.1, tlv8.DataType.FLOAT)
        entry2 = tlv8.Entry(1, 42.10001, tlv8.DataType.FLOAT)
        self.assertEqual(entry1, entry2)

    def test_equal(self):
        entry1 = tlv8.Entry(1, 42)
        entry2 = tlv8.Entry(1, 41)
        self.assertFalse(entry1 == entry2)
        self.assertNotEqual(entry1, entry2)

    def test_format_string(self):
        data = [
            tlv8.Entry(1, 3.141),
            tlv8.Entry(2, [
                tlv8.Entry(3, 'hello'),
                tlv8.Entry(4, 'world'),
            ]),
            tlv8.Entry(1, 2)
        ]
        print(tlv8.format_string(data))

    def test_format_string_error_1(self):
        self.assertRaises(ValueError, tlv8.format_string, {})

    def test_format_string_error_2(self):
        self.assertRaises(ValueError, tlv8.format_string, [1])

    def test_create_entry_error(self):
        self.assertRaises(ValueError, tlv8.Entry, 256, b'')



#
# Copyright 2020 Joachim Lusiardi
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import unittest
import enum
import binascii

import tlv8


class TestTLV8RealWorld(unittest.TestCase):
    def test_1(self):
        data = tlv8.EntryList([
            tlv8.Entry(1, b'W\x1ah\xac)\x04C\xfd\x84\xb36\t\xd1\x1bO\x83'),
            tlv8.Entry(3, tlv8.EntryList([
                tlv8.Entry(1, 0),
                tlv8.Entry(2, '192.168.178.222'),
            ]))
        ])
        encoded = tlv8.encode(data)
        decoded = tlv8.decode(
            encoded,
            {
                1: tlv8.DataType.BYTES,
                3: {
                    1: tlv8.DataType.UNSIGNED_INTEGER,
                    2: tlv8.DataType.STRING
                }
            }
        )
        self.assertEqual(data, decoded)
        encoded_2 = tlv8.encode(decoded)
        self.assertEqual(encoded, encoded_2)

    def test_2(self):
        class Foo(enum.IntEnum):
            Bar = 1
            Baz = 2

        data = tlv8.EntryList([
            tlv8.Entry(1, b'W\x1ah\xac)\x04C\xfd\x84\xb36\t\xd1\x1bO\x83'),
            tlv8.Entry(3, tlv8.EntryList([
                tlv8.Entry(1, Foo.Bar),
                tlv8.Entry(2, '192.168.178.222'),
            ]))
        ])
        encoded = tlv8.encode(data)
        decoded = tlv8.decode(
            encoded,
            {
                1: tlv8.DataType.BYTES,
                3: {
                    1: Foo,
                    2: tlv8.DataType.STRING
                }
            }
        )
        self.assertEqual(data, decoded)
        encoded_2 = tlv8.encode(decoded)
        self.assertEqual(encoded, encoded_2)

    def test_3(self):
        """
        Example for a broken decode from https://github.com/jlusiardi/homekit_python/issues/194.
        """
        data = '01ff0601040440371bf8c6cd6a7fb85e77a7b436a1b5300ed2023ca7e23f61303856a57358fd8ac03f288472776854765eb3' \
               'a2fcb4497c8b5497c29c88a574479030ec36b176ae05ff85c337879f0a4146a9cc089e0cb48ba3c9c21af0b206d493b224de' \
               'ee52ff0f9b1d8710db6531748ee6d1d66b8b4a6d0690670fb8f1233010d190c4ede1776cb10806eae66c78881647e82e9ba7' \
               'c806f52184c6f108275719cf425c4f8ea0e86c6534712343f88a1482c986e3dd252715872dee506520903c17d27f02ea8957' \
               '719c255631b78a9f2ecb7af0dc245b370cefef28f4652eebbe34afda0138039714665dd880559d1f2667294207892137820c' \
               'd80533d8c0b22601ffa49d1bdc1b641a33297fe59672a89d69391417c77e31283cd7f0d40920004d1bf1fc38357d9599ac2b' \
               '4d8ce3ac7ab8725a01500d198e94b00da80aac64ead393b266dcf9d4a07c05ff34548f7ebebd63f8a00ae2c82f6ee8ac6bcc' \
               'e0ab1030e9268c36714e2ec11c3bf21331129d62978e069dd087cbbdc31bdd6e0cf4ca825b91ab3c8b240de19aa097fc01cd' \
               '471e8c1b5598044d21be12b84c97a1d70e46681e5ecebbde1c33bae9bbd9b3ad41ba2aff8f1f952d0ef0cfcb8a674d5b4c7f' \
               '515ba94341334e86aac277920bd9080b9bf702e16671a3e41c0930beb8a552aefc28a3a9a7f2818e8fbb84c37ae10fb6c5d2' \
               '2e6ba9899e01f082381c9a3344ecbaf801ff85e33306ec9823e72ec4c93f9a45aa657b16f46757aaaf7c74daf35840e68749' \
               '42c132f4a639562920318b9f9867d8e5b0d50deac48c4e14842c91d565b0dd1fec667d092d123a4e1a05fff0b7070f184b4f' \
               '399532c0d1cc0aaea326efea765bc88ace048040a3e07a741e26ef55203bb3f76c075e3b6d20ede89a2eaa63e23376b0dff4' \
               'ef3a797df34a39d8130f60316e86cafcb264b0b570376ba911dc2bda031328ea1c915a724bc69ad2700623ed19c3ae75f946' \
               '4e3b1669adb916ad58e3252580911db2b535af3f2b2207aea880a0d24f1f759888bd5e25b6cf7b2e5ce825ab0fd943a8378a' \
               'eb12906e0965540af14dc0bfa3fae2eb5361992efaf56501ff6ed5c6e686a4af3c2fa121c810cb84cd5abac94f6d618af493' \
               '29e34fa613d2b758e3bc79eb03cb78328f9cd34df43566589615e42088681b4f69775350c9abf68c107d312b0f2421bb53cf' \
               '05ff50c14d6ff0da74b50d9b080c5c06175d66b24b35eb1e25f940170c0815a0ead23703b86da2103cd1b33021fd981d95c6' \
               'a32a3752dc903b0acba949d7d51a1bcabaebc52941bb25d558132feb1794481c0a5911e53553407a8771503d7673d4c3061a' \
               '4d2d41a2897fe507423509760fbe4847423a51155b99b67bf43c72958ce9409a459b5ce42e61309e96091411b256ec294fb0' \
               'f32782efc80d9d548f3cee4fdb21babd011e118238ec7545b24e5af74317a2670179930156512875653dce4e957e94a7596f' \
               'a7e2b533a3eacb9781634c79c094e2cbfcfc62128a25431f9b56cc40b6097614e7a4b08c32b3a7f2e471f55a295a9a06e5b0' \
               'e07dec2ad282842aa6f176052acd544cb5d67c206a3e38e80e32f560a57edda173892a39b021d616d8f2862a5d111e6610c1e7'
        data = binascii.unhexlify(data)
        resp_data = tlv8.decode(data, {1: tlv8.DataType.BYTES})
        self.assertIsNotNone(resp_data.first_by_id(1))
        self.assertEqual(len(resp_data.first_by_id(1).data), 510)

        resp_data = tlv8.decode(resp_data.first_by_id(1).data, {4: tlv8.DataType.BYTES, 6: tlv8.DataType.BYTES})
        self.assertIsNotNone(resp_data.first_by_id(4))
        print(resp_data.first_by_id(4).data)
        self.assertEqual(resp_data.first_by_id(4).data,
                         b'7\x1b\xf8\xc6\xcdj\x7f\xb8^w\xa7\xb46\xa1\xb50\x0e\xd2\x02<\xa7\xe2?a08V\xa5sX\xfd\x8a'
                         b'\xc0?(\x84rwhTv^\xb3\xa2\xfc\xb4I|\x8bT\x97\xc2\x9c\x88\xa5tG\x900\xec6\xb1v\xae')
        self.assertIsNotNone(resp_data.first_by_id(6))
        self.assertEqual(resp_data.first_by_id(6).data, b'\x04')



#
# Copyright 2020 Joachim Lusiardi
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import unittest

import tlv8


class TestTLV8EntryList(unittest.TestCase):
    def test_entrylist_init(self):
        with self.assertRaises(ValueError) as error_context:
            tlv8.EntryList('not a list')
        self.assertEqual(str(error_context.exception), 'No valid list: not a list')

        with self.assertRaises(ValueError) as error_context:
            tlv8.EntryList(['not an entry'])
        self.assertEqual(str(error_context.exception), 'Not a valid tlv8.Entry: not an entry')

    def test_entrylist_append(self):
        el = tlv8.EntryList()
        with self.assertRaises(ValueError) as error_context:
            el.append('not a list')
        self.assertEqual(str(error_context.exception), 'Not an tlv8.Entry: not a list')

    def test_entrylist_compare_empty(self):
        el_1 = tlv8.EntryList()
        el_2 = tlv8.EntryList()
        self.assertEqual(el_1, el_2)

    def test_entrylist_compare_not_a_list(self):
        el_1 = tlv8.EntryList()
        self.assertNotEqual(el_1, 'not a entry list')

    def test_entrylist_compare_1_equal(self):
        el_1 = tlv8.EntryList([tlv8.Entry(1, 2)])
        el_2 = tlv8.EntryList([tlv8.Entry(1, 2)])
        self.assertEqual(el_1, el_2)

    def test_entrylist_compare_1_not_equal(self):
        el_1 = tlv8.EntryList([tlv8.Entry(1, 2)])
        el_2 = tlv8.EntryList([tlv8.Entry(1, 3)])
        self.assertNotEqual(el_1, el_2)

    def test_entrylist_length(self):
        el = tlv8.EntryList()
        self.assertEqual(0, len(el))
        el.append(tlv8.Entry(1, 2))
        self.assertEqual(1, len(el))
        el.append(tlv8.Entry(1, 2))
        self.assertEqual(2, len(el))

    def test_entrylist_index(self):
        el = tlv8.EntryList()
        e1 = tlv8.Entry(1, 2)
        e2 = tlv8.Entry(3, 4)
        el.append(e1)
        el.append(e2)
        self.assertEqual(e2, el[1])

    def test_entrylist_decode(self):
        input_data = bytearray(b'\x02\x01\x23')
        result = tlv8.decode(input_data)
        self.assertIsInstance(result, tlv8.EntryList)
        self.assertEqual(result, tlv8.EntryList([tlv8.Entry(2, b'\x23')]))

    def test_entrylist_encode(self):
        el = tlv8.EntryList([tlv8.Entry(2, b'\x23')])
        result = el.encode()
        self.assertEqual(b'\x02\x01\x23', result)

    def test_entrylist_encode_same_sep_type(self):
        el = tlv8.EntryList([
            tlv8.Entry(2, b'\x23'),
            tlv8.Entry(2, b'\x42')
        ])
        result = el.encode(1)
        self.assertEqual(b'\x02\x01\x23\x01\x00\x02\x01\x42', result)

    def test_entrylist_assert_has(self):
        el = tlv8.EntryList([
            tlv8.Entry(2, b'\x23')
        ])
        with self.assertRaises(AssertionError) as error_context:
            el.assert_has(3, 'no bla bla')
        self.assertEqual(str(error_context.exception), 'no bla bla')

        el.assert_has(2, 'no bla bla')

    def test_entrylist_decode_nested(self):
        data = b'\x01\x04%\x06I@\x02\x0e\x03\x05hello\x04\x05world\x03\x01\x02'
        result = tlv8.decode(data, {
            1: tlv8.DataType.FLOAT,
            2: {
                3: tlv8.DataType.STRING,
                4: tlv8.DataType.STRING,
            },
            3: tlv8.DataType.INTEGER
        })
        self.assertIsInstance(result, tlv8.EntryList)
        self.assertIsInstance(result[1].data, tlv8.EntryList)

    def test_entrylist_by_id(self):
        el = tlv8.EntryList([
            tlv8.Entry(2, b'\x23'),
            tlv8.Entry(2, b'\x42')
        ])
        self.assertEqual(el.by_id(1), tlv8.EntryList())
        self.assertEqual(el.by_id(2), el)

    def test_entrylist_first_by_id(self):
        el = tlv8.EntryList([
            tlv8.Entry(2, b'\x23'),
            tlv8.Entry(2, b'\x42')
        ])
        self.assertEqual(el.first_by_id(1), None)
        self.assertEqual(el.first_by_id(2), el[0])

    def test_entrylist_format_string(self):
        el = tlv8.EntryList([
            tlv8.Entry(1, 1),
            tlv8.Entry(2, tlv8.EntryList([
                tlv8.Entry(4, 4),
                tlv8.Entry(5, 5)
            ])),
            tlv8.Entry(3, 3),
        ])
        result = tlv8.format_string(el)
        expected = """[
  <1, 1>,
  <2, [
    <4, 4>,
    <5, 5>,
  ]>,
  <3, 3>,
]"""
        self.assertEqual(result, expected)



#
# Copyright 2020 Joachim Lusiardi
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import unittest
from struct import pack

import tlv8


class TestTLV8(unittest.TestCase):
    def test_bytes_0length_bytes(self):
        entry = tlv8.Entry(23, b'', tlv8.DataType.BYTES)
        self.assertEqual(entry.encode(), b'\x17\x00')

    def test_bytes_0length_string(self):
        entry = tlv8.Entry(23, '', tlv8.DataType.STRING)
        self.assertEqual(entry.encode(), b'\x17\x00')

    def test_bytes_0length_autodetect(self):
        entry = tlv8.Entry(23, b'')
        self.assertEqual(entry.encode(), b'\x17\x00')

    def test_bytes_10bytes(self):
        entry = tlv8.Entry(23, b'0123456789', tlv8.DataType.BYTES)
        self.assertEqual(entry.encode(), b'\x17\x0a0123456789')

    def test_bytes_256bytes(self):
        data = b''
        for i in range(0, 256):
            data += pack('<B', i)
        entry = tlv8.Entry(23, data, tlv8.DataType.BYTES)
        expected_data = b'\x17\xff' + data[0:255] + b'\x17\x01' + data[255:]
        self.assertEqual(entry.encode(), expected_data)

    def test_bytes_257bytes(self):
        data = b''
        for i in range(0, 257):
            data += pack('<B', i % 256)
        entry = tlv8.Entry(23, data, tlv8.DataType.BYTES)
        expected_data = b'\x17\xff' + data[0:255] + b'\x17\x02' + data[255:]
        self.assertEqual(entry.encode(), expected_data)

    def test_string_0bytes(self):
        entry = tlv8.Entry(23, '', tlv8.DataType.STRING)
        self.assertEqual(entry.encode(), b'\x17\x00')

    def test_string_hello(self):
        entry = tlv8.Entry(23, 'hello', tlv8.DataType.STRING)
        self.assertEqual(entry.encode(), b'\x17\x05hello')

    def test_string_world(self):
        entry = tlv8.Entry(23, '🌍', tlv8.DataType.STRING)
        self.assertEqual(entry.encode(), b'\x17\x04\xf0\x9f\x8c\x8d')

    def test_float(self):
        entry = tlv8.Entry(23, 3.141, tlv8.DataType.FLOAT)
        self.assertEqual(entry.encode(), b'\x17\x04%\x06I@')

    def test_int_auto_length(self):
        entry = tlv8.Entry(23, 1, tlv8.DataType.INTEGER)
        self.assertEqual(entry.encode(), b'\x17\x01\x01')

    def test_int_overwrite_length(self):
        entry = tlv8.Entry(23, 1, tlv8.DataType.INTEGER, length=8)
        self.assertEqual(entry.encode(), b'\x17\x08\x01\x00\x00\x00\x00\x00\x00\x00')

    def test_unsigned_int_auto_length(self):
        entry = tlv8.Entry(23, 1, tlv8.DataType.UNSIGNED_INTEGER)
        self.assertEqual(entry.encode(), b'\x17\x01\x01')

    def test_unsigned_int_overwrite_length(self):
        entry = tlv8.Entry(23, 1, tlv8.DataType.UNSIGNED_INTEGER, length=8)
        self.assertEqual(entry.encode(), b'\x17\x08\x01\x00\x00\x00\x00\x00\x00\x00')

    def test_encode_different(self):
        data = [
            tlv8.Entry(23, b'23', tlv8.DataType.BYTES),
            tlv8.Entry(22, '23', tlv8.DataType.STRING)
        ]
        result = tlv8.encode(data)
        expected_data = data[0].encode() + data[1].encode()
        self.assertEqual(result, expected_data)

    def test_encode_same_autodetect(self):
        structure = [
            tlv8.Entry(1, 23),
            tlv8.Entry(2, 2345)
        ]
        result = tlv8.encode(structure)
        expected_data = b'\x01\x01\x17\x02\x02)\t'
        self.assertEqual(result, expected_data)

    def test_encode_same(self):
        data = [
            tlv8.Entry(23, b'23', tlv8.DataType.BYTES),
            tlv8.Entry(23, '23', tlv8.DataType.STRING)
        ]
        result = tlv8.encode(data)
        expected_data = data[0].encode() + b'\xff\x00' + data[1].encode()
        self.assertEqual(result, expected_data)

    def test_encode_same_set_sep_type(self):
        data = [
            tlv8.Entry(23, b'23', tlv8.DataType.BYTES),
            tlv8.Entry(23, '23', tlv8.DataType.STRING)
        ]
        result = tlv8.encode(data, 0)
        expected_data = data[0].encode() + b'\x00\x00' + data[1].encode()
        self.assertEqual(result, expected_data)

    def test_encode_same_set_sep_type_occurs(self):
        data = [
            tlv8.Entry(23, '42'),
            tlv8.Entry(23, '43')
        ]
        with self.assertRaises(ValueError) as error_context:
            tlv8.encode(data, 23)
        self.assertEqual(str(error_context.exception), 'Separator type id 23 occurs with list of entries!')

    def test_encode_3same(self):
        data = [
            tlv8.Entry(23, b'23', tlv8.DataType.BYTES),
            tlv8.Entry(23, '23', tlv8.DataType.STRING),
            tlv8.Entry(23, '23', tlv8.DataType.STRING)
        ]
        result = tlv8.encode(data)
        expected_data = \
            data[0].encode() + b'\xff\x00' + \
            data[1].encode() + b'\xff\x00' + \
            data[2].encode()
        self.assertEqual(result, expected_data)

    def test_encode_supported_rtp_configs(self):
        data = [
            tlv8.Entry(2, 0),
            tlv8.Entry(2, 1)
        ]
        result = tlv8.encode(data, separator_type_id=0x00)
        expected_data = \
            b'\x02\x01\x00' + \
            b'\x00\x00' + \
            b'\x02\x01\x01'
        self.assertEqual(result, expected_data)

    def test_encode_supported_video_stream_configuration(self):
        data = [
            tlv8.Entry(1, [  # video codec config
                tlv8.Entry(1, 0),  # h.264
                tlv8.Entry(2, [  # video codec param
                    tlv8.Entry(1, 0),  # Constrained Baseline Profile
                    tlv8.Entry(2, 0),  # Level 3.1
                    tlv8.Entry(3, 0),  # Packetization mode: Non-interleaved mode
                    tlv8.Entry(4, 1),  # CVO not enabled
                    tlv8.Entry(5, 10),  # CVO ID
                ]),
                tlv8.Entry(3, [  # video attributes
                    tlv8.Entry(1, 1280),  # width
                    tlv8.Entry(2, 800),  # height
                    tlv8.Entry(3, 30)  # fps
                ]),
                tlv8.Entry(3, [  # video attributes
                    tlv8.Entry(1, 640),  # width
                    tlv8.Entry(2, 480),  # height
                    tlv8.Entry(3, 30)  # fps
                ])
            ]),
            tlv8.Entry(1, [
                tlv8.Entry(1, 0),  # h.264
                tlv8.Entry(2, [  # video codec param
                    tlv8.Entry(1, 0),  # Constrained Baseline Profile
                    tlv8.Entry(2, 0),  # Level 3.1
                    tlv8.Entry(3, 0),  # Packetization mode: Non-interleaved mode
                    tlv8.Entry(4, 0),  # CVO not enabled
                ]),
                tlv8.Entry(3, [  # video attributes
                    tlv8.Entry(1, 1280),  # width
                    tlv8.Entry(2, 800),  # height
                    tlv8.Entry(3, 30)  # fps
                ]),
                tlv8.Entry(3, [  # video attributes
                    tlv8.Entry(1, 640),  # width
                    tlv8.Entry(2, 480),  # height
                    tlv8.Entry(3, 30)  # fps
                ])
            ])
        ]
        result = tlv8.encode(data)
        expected_data = \
            b'\x01\x30' \
            b'\x01\x01\x00' \
            b'\x02\x0f' \
            b'\x01\x01\x00' \
            b'\x02\x01\x00' \
            b'\x03\x01\x00' \
            b'\x04\x01\x01' \
            b'\x05\x01\n' \
            b'\x03\x0b' \
            b'\x01\x02\x00\x05' \
            b'\x02\x02 \x03' \
            b'\x03\x01\x1e' \
            b'\xff\x00' \
            b'\x03\x0b' \
            b'\x01\x02\x80\x02' \
            b'\x02\x02\xe0\x01' \
            b'\x03\x01\x1e' \
            b'\xff\x00' \
            b'\x01\x2d' \
            b'\x01\x01\x00' \
            b'\x02\x0c' \
            b'\x01\x01\x00' \
            b'\x02\x01\x00' \
            b'\x03\x01\x00' \
            b'\x04\x01\x00' \
            b'\x03\x0b' \
            b'\x01\x02\x00\x05' \
            b'\x02\x02 \x03' \
            b'\x03\x01\x1e' \
            b'\xff\x00' \
            b'\x03\x0b' \
            b'\x01\x02\x80\x02' \
            b'\x02\x02\xe0\x01' \
            b'\x03\x01\x1e'
        self.assertEqual(result, expected_data)

    def test_autodetection_of_types(self):
        data = [
            tlv8.Entry(1, 3.141),
            tlv8.Entry(2, [
                tlv8.Entry(3, 'hello'),
                tlv8.Entry(4, 'world')
            ]),
            tlv8.Entry(1, 2)
        ]
        result = tlv8.encode(data)
        expected_data = b'\x01\x04%\x06I@\x02\x0e\x03\x05hello\x04\x05world\x01\x01\x02'
        self.assertEqual(result, expected_data)

    def test_encode_error_1(self):
        self.assertRaises(ValueError, tlv8.encode, {})

    def test_encode_error_2(self):
        self.assertRaises(ValueError, tlv8.encode, [1])

    def test_encode_128bit_int(self):
        self.assertRaises(ValueError, tlv8.encode,
                          [tlv8.Entry(1, 85070591730234615865843651857942052864, tlv8.DataType.UNSIGNED_INTEGER)])

    def test_encode_128bit_signed_int(self):
        self.assertRaises(ValueError, tlv8.encode,
                          [tlv8.Entry(1, -85070591730234615865843651857942052864, tlv8.DataType.INTEGER)])

    def test_encode_64bit_int(self):
        result = tlv8.encode([tlv8.Entry(1, 4611686018427387904)])
        self.assertEqual(b'\x01\x08\x00\x00\x00\x00\x00\x00\x00@', result)

    def test_encode_32bit_int(self):
        result = tlv8.encode([tlv8.Entry(1, 1073741824)])
        self.assertEqual(b'\x01\x04\x00\x00\x00@', result)

    def test_encode_16bit_int(self):
        result = tlv8.encode([tlv8.Entry(1, 16384)])
        self.assertEqual(b'\x01\x02\x00@', result)

    def test_encode_8bit_signed_int(self):
        result = tlv8.encode([tlv8.Entry(1, -64, tlv8.DataType.INTEGER)])
        self.assertEqual(b'\x01\x01\xc0', result)

    def test_encode_8bit_unsigned_int(self):
        result = tlv8.encode([tlv8.Entry(1, 64, tlv8.DataType.UNSIGNED_INTEGER)])
        self.assertEqual(b'\x01\x01@', result)

    def test_encode_non_encodable(self):
        self.assertRaises(ValueError, tlv8.encode, [tlv8.Entry(1, {1: 2})])

    def test_encode_bytearray_autodetect(self):
        result = tlv8.encode([tlv8.Entry(1, bytearray(b'\x01'))])
        self.assertEqual(b'\x01\x01\x01', result)

    def test_encode_bytearray(self):
        result = tlv8.encode([tlv8.Entry(1, bytearray(b'\x01'), tlv8.DataType.BYTES)])
        self.assertEqual(b'\x01\x01\x01', result)



#
# Copyright 2020 Joachim Lusiardi
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import unittest
import json

import tlv8


class TestTLV8ToJson(unittest.TestCase):
    def test_entry_to_json(self):
        e = tlv8.Entry(1, 'hello')
        j = json.dumps(e, cls=tlv8.JsonEncoder)
        self.assertEqual('{"1": "hello"}', j)

    def test_entry_list_to_json(self):
        e1 = tlv8.Entry(1, 'hello')
        e2 = tlv8.Entry(2, 'world')
        el = tlv8.EntryList([e1, e2])
        j = json.dumps(el, cls=tlv8.JsonEncoder)
        self.assertEqual('[{"1": "hello"}, {"2": "world"}]', j)

    def test_string_to_json(self):
        j = json.dumps(True, cls=tlv8.JsonEncoder)
        self.assertEqual('true', j)
